#include "hparticlepool.h"
#include "hpidgeanttrackset.h"
#include "hgeantkine.h" 
#include "PParticle.h"
#include <TMath.h>
#include <TLorentzVector.h>
#include "hinvariantfactory.h"
#include "hmissingfactory.h"

using namespace std;

// ****************************************************************************
ClassImp(HParticlePool)

//-----------------------------------------------------------------------------
HParticlePool::HParticlePool() : 
	pSourceFile(0), pOutputFile(0), pInput(0), pFilter(0), ptrTarget(0), ptrBeam(0) 
{ 

}

//-----------------------------------------------------------------------------
HParticlePool::HParticlePool(const HFile& src) : 
	pSourceFile(&src), pOutputFile(0), pInput(new HInput(src)), pFilter(0), ptrTarget(0), ptrBeam(0)
{

}
	
//-----------------------------------------------------------------------------
HParticlePool::HParticlePool(const HFile& src, HFile& out) : 
	pSourceFile(&src), pOutputFile(&out), pInput(new HInput(src)), pFilter(0), ptrTarget(0), ptrBeam(0)
{

}
	
//-----------------------------------------------------------------------------
HParticlePool::HParticlePool(const HFile& src, HFile& out, const HFilter& fil) : 
	pSourceFile(&src), pOutputFile(&out), pInput(new HInput(src)), pFilter(&fil), ptrTarget(0), ptrBeam(0)
{

}

//-----------------------------------------------------------------------------
HParticlePool::HEntity::HEntity(HParticlePool& ref, std::string name, std::string type) : name(name), ptrPF(0), ptrHN(0)
{
   if (type=="invariant") ptrPF = new HInvariantFactory(ref, name);
   if (type=="missing") ptrPF = new HMissingFactory(ref, name);
   ptrHN = new HNtuple(name.c_str(), name.c_str());
}

//-----------------------------------------------------------------------------
HParticlePool::HEntity::~HEntity()
{
   if (ptrPF) delete ptrPF;
   if (ptrHN) delete ptrHN;
}


	
//-----------------------------------------------------------------------------
HParticlePool::~HParticlePool() {
   if (ptrTarget) delete ptrTarget;
   if (ptrBeam) delete ptrBeam;
   std::for_each( objectList.begin(), objectList.end(), DelListObj() );
}

//-----------------------------------------------------------------------------
void HParticlePool::setSource(const HFile& ref)
{
   pSourceFile = &ref;
   pInput = new HInput(ref);
}

//-----------------------------------------------------------------------------
void HParticlePool::setOutput(HFile& ref)
{
   pOutputFile = &ref;
}

//-----------------------------------------------------------------------------
void HParticlePool::setFilter(const HFilter& ref)
{
   pFilter = &ref;
}

//-----------------------------------------------------------------------------
int HParticlePool::count( EParticle n )
{
   return particle.count( n );
}

//-----------------------------------------------------------------------------
MultiParticleIterPair HParticlePool::equal_range( EParticle n )
{
   return particle.equal_range( n );
}

//-----------------------------------------------------------------------------
MultiParticleIter HParticlePool::lower_bound( EParticle n )
{
   return particle.lower_bound( n );
}

//-----------------------------------------------------------------------------
MultiParticleIter HParticlePool::upper_bound( EParticle n )
{
	   return particle.upper_bound( n );
}

//-----------------------------------------------------------------------------
void HParticlePool::target(const char* part, double px, double py, double pz)
{
   if (ptrTarget) delete ptrTarget;
   ptrTarget = new HParticle(part, px, py, pz);
   ptrTarget->setAcc(1.0);
   ptrTarget->setEff(1.0);
}

//-----------------------------------------------------------------------------
void HParticlePool::beam(const char* part, double px, double py, double pz)
{
   if (ptrBeam) delete ptrBeam;
   ptrBeam = new HParticle(part, px, py, pz);
   ptrBeam->setAcc(1.0);
   ptrBeam->setEff(1.0);
}


//-----------------------------------------------------------------------------
void HParticlePool::invariant(const char* name)
{
   if (false == isObject(name))
   {
      objectList.push_back( new HEntity( *this, name, "invariant") );
   }
}


//-----------------------------------------------------------------------------
void HParticlePool::missing(const char* name)
{
   if (false == isObject(name))
   {
      objectList.push_back( new HEntity( *this, name, "missing") );
   }
}

//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, EParticle p1)
{
   HEntity* ptr = getObject( name );
   if (ptr)
   {
      ptr->ptrPF->add( p1 );
   }
}

//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, const char* p1)
{
   add( name, convertName(p1) );
}


//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, EParticle p1, EParticle p2)
{
   HEntity* ptr = getObject( name );
   if (ptr)
   {
      ptr->ptrPF->add( p1, p2 );
   }
}

//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, const char* p1, const char* p2)
{
   add( name, convertName(p1), convertName(p2) );
}


//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, EParticle p1, EParticle p2, EParticle p3)
{
   HEntity* ptr = getObject( name );
   if (ptr)
   {
      ptr->ptrPF->add( p1, p2, p3 );
   }
}

//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, const char* p1, const char* p2, const char* p3)
{
   add( name, convertName(p1), convertName(p2), convertName(p3) );
}

//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, EParticle p1, EParticle p2, EParticle p3, EParticle p4)
{
   HEntity* ptr = getObject( name );
   if (ptr)
   {
      ptr->ptrPF->add( p1, p2, p3, p4 );
   }
}

//-----------------------------------------------------------------------------
void HParticlePool::add(const char* name, const char* p1, const char* p2, const char* p3, const char* p4)
{
   add( name, convertName(p1), convertName(p2), convertName(p3), convertName(p4) );
}

//-----------------------------------------------------------------------------
bool HParticlePool::isObject(const char* name)
{
   if (objectList.end() == std::find_if( objectList.begin(), objectList.end(), std::bind2nd( IsName(), name ) ))
   {
      return false;
   }
return true;
}

//-----------------------------------------------------------------------------
HParticlePool::HEntity* HParticlePool::getObject(const char* name)
{
   if ( isObject(name) )
   {
      return *std::find_if( objectList.begin(), objectList.end(), std::bind2nd( IsName(), name ));
   }
return 0;
}


//-----------------------------------------------------------------------------
EParticle HParticlePool::convertName(const char* name)
{
   string tmp_name(name);
   if (tmp_name == "ep") return eEP;
   else if (tmp_name == "em") return eEM;
   else if (tmp_name == "pip") return ePIP;
   else if (tmp_name == "pim") return ePIM;
   else if (tmp_name == "p") return eP;
}

//-----------------------------------------------------------------------------
EParticle HParticlePool::convertId(int n)
{
   if (n == 2) return eEP;
   else if (n == 3) return eEM;
   else if (n == 8) return ePIP;
   else if (n == 9) return ePIM;
   else if (n == 14) return eP;
}

//-----------------------------------------------------------------------------
void HParticlePool::loop()
{
   if (pSourceFile)
   {
      int sourceSize = pSourceFile->size();
      for (int i=0; i<sourceSize; ++i)
      {
         pInput->set( i );
         for (int j=0; j<pInput->nentries(); ++j)
         {
            pInput->getEntry(j);
            // ----------------------------------
            particle.clear();
            particleNum.clear();
            for (int k=0; k<pInput->nparticles(); ++k)
            {
               HParticle tmp = pInput->read(k);
               // ----------------------------------
               // here some filtering and smearing
               // ----------------------------------
	       pFilter->filter( tmp );
               particle.insert(pair<EParticle, HParticle>( convertId(tmp.getId()), tmp));
               particleNum[ convertId(tmp.getId()) ]++;
            }
            for (ParticleNum::iterator it = particleNum.begin(); it != particleNum.end(); ++it)
            {
               //     std::cout << "CZASTKA: " << it->first << " jest " << it->second << " razy." << std::endl;
            }
            for (std::list<HEntity*>::iterator listIt = objectList.begin(); listIt != objectList.end(); ++listIt)
            {
                (*listIt)->ptrPF->combine();
		for (int k=0; k<(*listIt)->ptrPF->size(); ++k)
		{
                   dump( *(*listIt)->ptrPF->get(k), *(*listIt)->ptrHN );
		}
            }

         }
      }
   }
}

//-----------------------------------------------------------------------------
void HParticlePool::dump(HNParticle& refP, HNtuple& ntuple)
{
   ntuple["id"] = refP.getId();
   ntuple["mass"] = refP.getMass();
   ntuple["acc"] = refP.getAcc();
   ntuple["eff"] = refP.getEff();
   ntuple["p"] = refP.getP();
   ntuple["px"] = refP.getPx();
   ntuple["py"] = refP.getPy();
   ntuple["pz"] = refP.getPz();
   ntuple["theta"] = refP.getTheta();
   ntuple["phi"] = refP.getPhi();

   ntuple.fill();
}

//-----------------------------------------------------------------------------
void HParticlePool::write()
{
   pOutputFile->getOutput();
   for (std::list<HEntity*>::iterator listIt = objectList.begin(); listIt != objectList.end(); ++listIt)
   {
       for (int k=0; k<(*listIt)->ptrPF->size(); ++k)
       {
          (*listIt)->ptrHN->Write(); 
       }
   }
}
